import { configs as tseConfig, parser as tseParser, plugin as tsePlugin } from "typescript-eslint";
import tsdocPlugin from "eslint-plugin-tsdoc";
import importPlugin from "eslint-plugin-import";

/**
 * @typedef {import("eslint").Linter.Config} Config
 */

/**
 * @param {Config} config
 * @returns {Config}
 */
function defineConfig(config) {
	return config;
}

function flatten(result, it) {
	return {
		...result,
		...it,
		languageOptions: {
			...result.languageOptions,
			...it.languageOptions,
		},
		plugins: {
			...result.plugins,
			...it.plugins,
		},
		rules: {
			...result.rules,
			...it.rules,
		},
	};
}

/** @type {Config} */
const strict = tseConfig.strict.reduce(flatten, {});

/** @type {Config} */
const stylistic = tseConfig.stylistic.reduce(flatten, {});

/** @type {Config} */
const importConfig = importPlugin.flatConfigs.typescript;

export default defineConfig({
	languageOptions: {
		parser: tseParser,
		sourceType: "module",
	},
	plugins: {
		"@typescript-eslint": tsePlugin,
		tsdoc: tsdocPlugin,
	},
	settings: {
		...importConfig.settings,
	},
	rules: {
		...strict.rules,
		...stylistic.rules,
		...importConfig.rules,

		/* typescript handles the return types */
		"consistent-return": "off",

		"tsdoc/syntax": "error",

		/* prefer T[] for simple types, Array<T> for complex types (unions, etc) */
		"@typescript-eslint/array-type": ["error", { default: "array-simple" }],

		/* allow getters which returns a literal */
		"@typescript-eslint/class-literal-property-style": "off",

		"@typescript-eslint/explicit-function-return-type": [
			"error",
			{
				allowExpressions: true,
			},
		],
		"@typescript-eslint/explicit-member-accessibility": "error",
		"@typescript-eslint/no-inferrable-types": "off",

		/* allow function(this: void, ...) */
		"@typescript-eslint/no-invalid-void-type": [
			"error",
			{
				allowAsThisParameter: true,
			},
		],

		/* allow constructs such as `unknown | null`, while `unknown` does override
		 * `null` it can still serve as a self-documenting type to signal that
		 * `null` has a special meaning. It also helps when the type is to be
		 * replaced with a better type later. */
		"@typescript-eslint/no-redundant-type-constituents": "off",

		"@typescript-eslint/no-unused-vars": [
			"error",
			{
				ignoreRestSiblings: true,
				argsIgnorePattern: "^_",
			},
		],

		/* allow overloading only if the parameters have different names */
		"@typescript-eslint/unified-signatures": [
			"error",
			{
				ignoreDifferentlyNamedParameters: true,
			},
		],
	},
});
